import numpy as np

def gauss_seidel(A, b, tol=1e-10, max_iter=1000):
    """
    Resuelve sistemas de ecuaciones lineales Ax = b usando el metodo de gauss-seidel

    Parametros:
        A: Matriz de coeficientes.
        b: Vector de terminos independientes.
        tol: Tolerancia para la convergencia.
        max_iter: Maximo numero de iteraciones.

    Retorna:
        x: Solucion aproximada.
    """

    n = len(b)
    x = np.zeros(n)

    for k in range(max_iter):
        x_prev = x.copy()

        for i in range(n):
            x[i]= (b[i] - sum(A[i,j] * x[j] for j in range(n) if j != i)) / A[i,i]

        if np.linalg.norm(x - x_prev) < tol:   #verificamos la convergencia
            return x
        
    print(f"No se pudo converger en {max_iter} iteraciones.")
    return x






import numpy as np

def jacobi(A, b, tol=1e-10, max_iter=1000):
    """
    Resuelve sistemas de ecuaciones lineales Ax = b usando el metodo de Jacobi

    Parametros:
        A: Matriz de coeficientes.
        b: Vector de terminos independientes.
        tol: Tolerancia para la convergencia.
        max_iter: Maximo numero de iteraciones.

    Retorna:
        x: Solucion aproximada.
    """
    n = len(b)
    x = np.zeros(n)
    x_new = np.zeros(n)


    for k in range(max_iter):
        for i in range(n):
            x_new[i]= (b[i] - sum(A[i,j] * x[j] for j in range(n) if j != i))/A[i,i]

        if np.linalg.norm(x_new - x) < tol:
            print(f"Total de iteraciones necesitadas: {k}")
            return x_new
        
        x = x_new.copy()
    
    print(f"No se pudo converger en {max_iter} iteraciones.")
    return x



import numpy as np
def eliminacion_gauss(A, b):
    """
    Implementación del método de eliminación de Gauss para resolver Ax = b
    """
    n = len(b)
    Ab = np.hstack((A, b.reshape(-1, 1)))

    for i in range(n):
        max_row = np.argmax(np.abs(Ab[i:, i])) + i
        Ab[[i, max_row]] = Ab[[max_row, i]]

        for j in range(i + 1, n):
            factor = Ab[j, i] / Ab[i, i]
            Ab[j] -= factor * Ab[i]

    x = np.zeros(n)
    for i in range(n - 1, -1, -1):
        x[i] = (Ab[i, -1] - np.dot(Ab[i, i + 1:n], x[i + 1:])) / Ab[i, i]

    return x


import numpy as np

def descomposicion_lu(A):
    """
    Realiza la descomposición LU de una matriz cuadrada A usando el algoritmo.
    
    Parametros:
      A (np.array): Matriz de coeficientes (n x n).

    Retorna:
      L (np.array): Matriz triangular inferior con 1's en la diagonal.
      U (np.array): Matriz triangular superior.
    
    Nota: Se asume que A es no singular y que no se requiere pivoteo.
    """
    A = A.astype(float)
    n = A.shape[0]
    
    # Inicializamos L y U
    L = np.eye(n, dtype=float)
    U = np.zeros((n, n), dtype=float)
    
    # Algoritmo de Doolittle
    for i in range(n):
        # Calcular la fila i de U
        for j in range(i, n):
            sum_u = 0.0
            for k in range(i):
                sum_u += L[i, k] * U[k, j]
            U[i, j] = A[i, j] - sum_u

        # Calcular la columna i de L (para elementos debajo de la diagonal)
        for j in range(i + 1, n):
            sum_l = 0.0
            for k in range(i):
                sum_l += L[j, k] * U[k, i]
            L[j, i] = (A[j, i] - sum_l) / U[i, i]
    
    return L, U

def lu_solve(A, b):
    """
    Resuelve el sistema Ax = b usando la descomposición LU.
    
    Parámetros:
      A (np.array): Matriz de coeficientes (n x n).
      b (np.array): Vector de términos independientes (n,).
      
    Retorna:
      x (np.array): Solución del sistema.
    """
    L, U = descomposicion_lu(A)
    n = A.shape[0]
    
    # Resolución del sistema L y = b mediante sustitución progresiva:
    y = np.zeros(n, dtype=float)
    for i in range(n):
        sum_y = 0.0
        for j in range(i):
            sum_y += L[i, j] * y[j]
        y[i] = b[i] - sum_y
        
    # Resolución del sistema U x = y mediante sustitución regresiva:
    x = np.zeros(n, dtype=float)
    for i in range(n - 1, -1, -1):
        sum_x = 0.0
        for j in range(i + 1, n):
            sum_x += U[i, j] * x[j]
        x[i] = (y[i] - sum_x) / U[i, i]
        
    return x


import numpy as np

def cramer(A, b):
    """
    Resuelve el sistema Ax = b utilizando la regla de Cramer.
    
    Parámetros:
    - A: np.array o lista de listas que representa la matriz de coeficientes (n x n).
    - b: np.array o lista que representa el vector de términos independientes (n).
    
    Retorna:
    - x: np.array conteniendo la solución del sistema.
    
    Lanza:
    - ValueError: Si el determinante de A es cero, indicando que el sistema no tiene solución única.
    """
    A = np.array(A, dtype=float)
    b = np.array(b, dtype=float)
    
    n = len(b)
    D = np.linalg.det(A)  # Determinante de A

    if np.isclose(D, 0):
        raise ValueError("El determinante de A es cero. El sistema no tiene solución única.")

    x = np.zeros(n)

    for k in range(n):
        Ak = A.copy()
        Ak[:, k] = b  # Reemplaza la columna k por el vector b
        Dk = np.linalg.det(Ak)
        x[k] = Dk / D

    return x


from .eliminacion_gauss import eliminacion_gauss
from .gauss_jordan import gauss_jordan
from .crammer import crammer
from .descomposicion_lu import descomposicion_lu
from .jacobi import jacobi
from .gauss_seidel import gauss_seidel
from .biseccion import biseccion

__all__ = [
    "eliminacion_gauss",
    "gauss_jordan",
    "crammer",
    "descomposicion_lu",
    "jacobi",
    "gauss_seidel",
    "biseccion"
]



import numpy as np

def gauss_jordan(coefficients, ind_terms):
    """
    Resuelve un sistema de ecuaciones lineales Ax = b usando el método de Gauss-Jordan.
    """
    coefficients = np.array(coefficients, dtype=float)
    ind_terms = np.array(ind_terms, dtype=float)

    n = len(ind_terms)
    equations = np.hstack([coefficients, ind_terms.reshape(-1, 1)])

    for i in range(n):
        max_row = max(range(i, n), key=lambda r: abs(equations[r, i]))
        equations[[i, max_row]] = equations[[max_row, i]]

        if np.isclose(equations[i, i], 0): 
            return None

        equations[i] /= equations[i, i]

        for j in range(n):
            if i != j:
                factor = equations[j, i]
                equations[j] -= factor * equations[i]

    return equations[:, -1]





import math

def biseccion(f,a,b,tol=1e-6,max_iter=1000):
    """
    Resuelve ecuaciones lineales, encontrando la raiz. f(x) = 0 en [a,b]
    Parametros:
        f: funcion a evaluar
        a: valor menor
        b: valor mayor
        tol: tolerancia
        max_iter: maximo de iteraciones
    Retorna:
        float: Raiz aproximada
        int: Numero de iteraciones.

    """
    
    if(f(a)*f(b)) >= 0:
        raise ValueError("Error, la funcion no cambia de signo en [a,b]")
    iteraciones = 0
    while (b-a)/2 > tol and iteraciones < max_iter:
        c = (a + b) / 2
        if (f(c) == 0):
            return c, iteraciones
        elif f(a) * f(c) < 0:
            b = c #la raiz se encuentra entre a y c
        else:
            a = c #la raiz se encuentra entre c y b
        iteraciones += 1

    return (a + b) / 2, iteraciones


